/****************************************
 * Gestion de l'interface avec le shell
 * init_pipe() dmarre le shell et cre les deux pipes
 * envoieshell() envoie une commande  sh (sans '\n' au bout)
 * recoitshell() reoit une ligne de la rponse (ou NULL aprs la dernire)
 * execshell() excute une commande shell sans attendre la rponse
 * erreur() provoque une erreur gnrale
 * *************************************/

#include "prehtml.h"

/* fichiers inclus uniquement dans cette partie, pour cette raison
 * ils ne sont pas dans prehtml.h */
#include <signal.h>
#include <wait.h>

FILE *pin=NULL,*pout=NULL;
static int result_fork;
static string envoyee;

void erreur(const char *format, ...){
  va_list ap;
  va_start(ap,format);
  dit_ou();
  fprintf(stderr,"prehtml: ");
  vfprintf(stderr,format,ap);
  putc('\n',stderr);
  exit(1);
}

void warning(const char *format, ...){
  va_list ap;
  va_start(ap,format);
  dit_ou();
  fprintf(stderr,"prehtml: attention: ");
  vfprintf(stderr,format,ap);
  putc('\n',stderr);
}

static vector<string> listetemp;

void creetemp(const char *nom){ /* signale la cration d'un fichier temporaire,
			   * pour permettre le nettoyage si ncessaire */
  listetemp.push_back(nom);
}

void fin_urgente(int numsignal){
  char nom[10];
  switch(numsignal){
   case SIGALRM: strcpy(nom,"SIGALRM");break;
   case SIGHUP: strcpy(nom,"SIGHUP");break;
   case SIGINT: strcpy(nom,"SIGINT");break;
   case SIGQUIT: strcpy(nom,"SIGQUIT");break;
   case SIGABRT: strcpy(nom,"SIGABT");break;
   default: sprintf(nom,"n%-5d",numsignal);
  }
  if (numsignal)
    warning("Interruption reue: %s\n"
	  "commande ventuellement en cours de traitement: \n"
	  "   <%s %s>\n"
	  "Nettoyage des fichiers temporaires",
	  nom,(fonction.size()?fonction.c_str():"[aucune]"),chaine.c_str());
  else fprintf(stderr,"Nettoyage des fichiers temporaires\n\n");
  /* la fonction suivante est une extension GNU !!!
   elle sert  fermer les fichiers avant de les refermer;
   il semble que a marche quand mme si on ne la met pas...*/
  fcloseall();
  for(size_t n=0; n<listetemp.size(); n++) /* on a cr des fichiers temporaires*/
    unlink(listetemp[n].c_str());
  listetemp.clear();
  if (numsignal){
    /* je remets le comportement pas dfaut, et je relance le signal */
    signal(numsignal, SIG_DFL);
    raise(numsignal);
  }
}
  

static int alarmstop=0;
void alarme(int numsignal){
  if (alarmstop){
    fprintf(stderr,"shell: la commande:\n     \"%s\"\n"
	 " ne rpond pas: peut-tre ceci vient-il de la balise:\n"
	 "     <%s %s>\n"
	 "Vrifiez les guillemets incomplets, crochets, etc...\n\n",
	 envoyee.c_str(),fonction.c_str(),chaine.c_str());
    fin_urgente(0);
    exit(1);
  }
  else
    warning("shell:  la commande:\n     \"%s\"\n"
	    " semble durer un peu longtemps: est-ce cette balise?\n"
	    "     <%s %s>",
	 envoyee.c_str(),fonction.c_str(),chaine.c_str());
}

void init_pipe(void){
  int fd[4];
  pipe(fd);
  pipe(fd+2);
  result_fork=fork();
  if (result_fork==-1) erreur("Le fork n'a pas march");
  if (result_fork==0){ // process fils
    dup2(fd[0],0); // stdin
    dup2(fd[3],1); // stdout
    setlinebuf(stdin);
    setlinebuf(stdout);
    execlp("sh","sh",NULL);
    erreur("exec sh n'a pas march");
  }
  else{
    /* ... car si on ne les ouvre pas dans le mme ordre ici, a bloque */
    pin=fdopen(fd[2],"rb");
    pout=fdopen(fd[1],"wb");
    setlinebuf(pin);
    setlinebuf(pout);
    // on vrifie que a fonctionne
    fputs("echo\n",pout);
    fflush(pout);
    while(fgetc(pin)!='\n');
  }
  if (signal (SIGALRM,alarme)==SIG_ERR ||
      signal(SIGHUP,fin_urgente)==SIG_ERR ||
      signal(SIGINT,fin_urgente)==SIG_ERR ||
      signal(SIGQUIT,fin_urgente)==SIG_ERR ||
      signal(SIGABRT,fin_urgente)==SIG_ERR )
    fprintf(stderr,
	    "prehtml: attention: des interruptions n'ont pas pu tre dtournes\n");
}

void fin_pipe(void){
  if (!pout) erreur("fin_pipe(): pout dj ferm");
  fclose(pout);
  if (!pin) erreur("fin_pipe(): pin dj ferm");
  fclose(pin);
  pout=pin=NULL;
  kill(result_fork,SIGTERM);
  wait(NULL);
}
  

void envoieshell(string &s){  // envoie une commande  sh
  if (shell)
    fprintf(stderr,"prehtml: shell: > \"%s\"\n",s.c_str());
  char s2[80+1];
  if (!pout) erreur("envoieshell: pipe out non ouvert!");
  envoyee=s;
  fprintf(pout,"echo COMMANDE_TEST;%s\necho %s\n",s.c_str(),CODE_FIN_SHELL);
  fflush(pout);
  alarmstop=1; /* si timeout ici, c'est la fin: le shell ne se lance pas */
  alarm(3); // alarme dans 3 secondes !
  fgets(s2,80,pin);
  alarm(30); // OK, j'ai reu quelques chose
  alarmstop=0; /* si timeout ici, c'est peut-tre que la commande shell est
		* simplement trs longue: simple warning */
  if (strcmp(s2,"COMMANDE_TEST\n"))
    erreur("shell: Avant mme d'envoyer \"%s\",\n"
	   "  je reois \"%s\". Erreur interne ?",
	   s.c_str(),s2);
}


int recoitshell(string &s, int retireEOL){ // rcupre la rponse de sh et met une ligne
                            // dans s; renvoie 0 si c'est fini, 1 sinon
  s.clear();
  char lignerecue[80+1];  


  static int fin_ligne_otee=0;
  int fin_prec;
  int i;
  
  if (retireEOL) fin_ligne_otee=0;
  if (fin_ligne_otee)
    s="\n";
  fin_ligne_otee=0;
  fin_prec=s.size();
  while(!feof(pin)){
    fgets(lignerecue,80,pin);
    i=strlen(lignerecue);
    if (i) if (lignerecue[i-1]=='\n') {lignerecue[i-1]='\0';fin_ligne_otee=1;}
    s += lignerecue;
    if (fin_ligne_otee) break;
  }
  if (!s.compare(fin_prec,strlen(CODE_FIN_SHELL),CODE_FIN_SHELL)) {
    fin_ligne_otee=0; // le dernier \n doit bien tre t
    alarm(0); // ma commande shell est finie, donc le timeout aussi
    return 0;
  }
  if (shell)
    fprintf(stderr,"prehtml: shell: < \"%s\"\n",s.c_str());
  return 1;
}

void execshell(string &s){ // envoie une commande  sh mais nglige la rponse
  string bidon;
  envoieshell(s);
  while (recoitshell(bidon,0));
}
